  {      LDAPAdmin - Xml.pas
  *      Copyright (C) 2005 Alexander Sokoloff
  *
  *      Author: Alexander Sokoloff
  *
  *
  * This file is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This file is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

    @Version 2006/03/18  v1.0.0. Initial revision.
    @Version 2006/09/13  v1.1.0. Added TXmlNode.CoptTo function
  }

unit Xml;

interface

uses windows, sysutils, Classes, Contnrs;

type
  TTagType=(tt_Open, tt_Close, tt_Single);

  TXmlNode=class
  private
    FName:      string;
    FAttrs:     TStringList;
    FNodes:     TObjectList;
    FParent:    TXmlNode;
    FContent:   string;
    FCaseSens:  boolean;
    function    GetNodes(Index: integer): TXmlNode;
    function    GetCount: integer;
    procedure   SetParent(const Value: TXmlNode);
    function    GetCaseSens: boolean;
  public
    constructor Create(AParent: TXmlNode); reintroduce;
    destructor  Destroy; override;
    property    Attributes: TStringList read FAttrs;
    property    Nodes[Index: integer]: TXmlNode read GetNodes; default;
    property    Count: integer read GetCount;
    property    Name: string read FName write FName;
    property    Parent: TXmlNode read FParent write SetParent;
    property    Content: string read FContent write FContent;
    function    Add(const AName: string=''; const AContent: string=''; const AAttributes: TstringList=nil): TXmlNode;
    function    Clone(const Recurse: boolean): TxmlNode;
    procedure   Delete(Index: integer);
    function    NodeByName(AName: string; CaseSensitive: boolean=true): TXmlNode;
    property    CaseSensitive: boolean read GetCaseSens;
   end;

  TXmlTree=class
  private
    FRoot:      TXmlNode;
    function    GetNextTag(Stream: TStream; var PrevContent, TagName, Attrs: string; var TagType: TTagType): boolean;
    procedure   ParseAttributes(S: string; Attributes: TStringList);
    function    ClearContent(S: string): string;
    procedure   ParsePath(const Path: string; Result: TStrings);
    function    DoGetNode(var Path: TStrings): TXmlNode;
    procedure   SetCaseSens(const Value: boolean);
    function    GetCaseSens: boolean;
  public
    constructor Create; reintroduce;
    destructor  Destroy; override;
    property    Root: TXmlNode read FRoot;
    procedure   LoadFromStream(const Stream: TStream);
    procedure   SaveToStream(const Stream: TStream);
    procedure   LoadFromFile(const FileName: string);
    procedure   SaveToFile(const FileName: string);
    function    ByPath(const APath: string): TXmlNode;
    function    Exist(const Path: string): boolean;
    property    CaseSensitive: boolean read GetCaseSens write SetCaseSens;
  end;

implementation

const
  BAD_XML_DOCUMENT='Not well formed XML.';
  XML_BAD_CLOSE_TAG=BAD_XML_DOCUMENT+#10+'The "%s" is expected, but the "%s" is received';
  XML_UNEXPECTED_CLOSE_TAG=BAD_XML_DOCUMENT+#10+'Unexpected closing tag "%s"';
  TAB='  ';//#9;
  CRLF=#13#10;

{ TXmlNode }

constructor TXmlNode.Create(AParent: TXmlNode);
begin
  inherited Create;
  FAttrs:=TStringList.Create;
  FNodes:=TObjectList.Create;
  FParent:=AParent;
  if FParent<>nil then FParent.FNodes.Add(self);
end;

destructor TXmlNode.Destroy;
begin
  FAttrs.Free;
  FNodes.Free;
  inherited;
end;

function TXmlNode.GetNodes(Index: integer): TXmlNode;
begin
  result:=TXmlNode(FNodes[Index]);
end;

function TXmlNode.GetCount: integer;
begin
  result:=FNodes.Count;
end;

function TXmlNode.Add(const AName, AContent: string; const AAttributes: TstringList): TXmlNode;
begin
  result:=TXmlNode.Create(self);
  result.Name:=AName;
  result.Content:=AContent;
  if AAttributes<>nil then Attributes.Assign(AAttributes);
end;

function TXmlNode.NodeByName(AName: string; CaseSensitive: boolean=true): TXmlNode;
type
  TCompareProc=function(const AText, AOther: string): Boolean;
var
  i: integer;
  proc: TCompareProc;
begin
  result:=nil;
  if CaseSensitive then proc:=AnsiSameStr
  else proc:=AnsiSameText;

  for i:=0 to Count-1 do begin
    if proc(Nodes[i].Name, AName) then begin
      result:=Nodes[i];
      exit;
    end;
  end;
end;

procedure TXmlNode.Delete(Index: integer);
begin
  FNodes.Delete(Index);
end;

procedure TXmlNode.SetParent(const Value: TXmlNode);
begin
  if FParent<>nil then FParent.FNodes.Extract(self);
  FParent := Value;
  if FParent<>nil then FParent.FNodes.Add(self);
end;

function TXmlNode.GetCaseSens: boolean;
begin
  if FParent=nil then result:=FCaseSens
  else result:=FParent.CaseSensitive;
end;

function TXmlNode.Clone(const Recurse: boolean): TxmlNode;
  function DoAdd(Src, Dest: TXmlNode): TXmlNode;
  var
    i: integer;
  begin
    result:=TXmlNode.Create(Dest);
    result.Name:=Src.Name;
    result.Content:=Src.Content;
    result.Attributes.Assign(Src.Attributes);
    if Recurse then
      for i:=0 to Src.Count-1 do DoAdd(Src.Nodes[i], result);
  end;

begin
  result:=DoAdd(self, nil);
end;


{ TXmlTree }

constructor TXmlTree.Create;
begin
  inherited;
  FRoot:=TXmlNode.Create(nil);
end;

destructor TXmlTree.Destroy;
begin
  FRoot.Free;
  inherited;
end;

function TXmlTree.GetNextTag(Stream: TStream; var PrevContent, TagName, Attrs: string; var TagType: TTagType): boolean;
type
  TState=(tsContent, tsTag, tsAttrs, tsEnd, tsSkip, tsComment);
var
  c: char;
  State: TState;
  quota: boolean;
  CommentState: byte;
begin
  result:=false;
  PrevContent:='';
  TagName:='';
  attrs:='';
  TagType:=tt_Open;
  State:=tsContent;
  quota:=false;
  CommentState:=0;
  while (Stream.Position<Stream.Size) do begin
    Stream.ReadBuffer(c, sizeof(c));
    case State of
      //////////////////////////////////////////////////////////////////////////
      tsContent:  case c of
                    '<':  State:=tsTag;
                    else  PrevContent:=PrevContent+c;
                  end;
      //////////////////////////////////////////////////////////////////////////
      tsTag:      case c of
                    '''',
                    '"': if TagName='' then quota:=true
                         else State:=tsAttrs;
                    '/':  if TagName='' then TagType:=tt_Close
                          else TagType:=tt_Single;
                    '>':  begin
                            result:=true;
                            break;
                          end;
                    ' ':  if quota then TagName:=TagName+c
                          else State:=tsAttrs;
                    '?':  if TagName='' then State:=tsSkip;
                    '-':  if TagName='!-' then begin
                            CommentState:=0;
                            TagName:='';
                            State:=tsComment;
                          end
                          else TagName:=TagName+c;
                    else  TagName:=TagName+c;
                  end;
     //////////////////////////////////////////////////////////////////////////
     tsAttrs:     case c of
                    '>':  begin
                            result:=true;
                            break;
                          end;
                    else  attrs:=attrs+c;
                  end;
     //////////////////////////////////////////////////////////////////////////
     tsSkip:      case c of
                    '>':  State:=tsContent;
                  end;
     //////////////////////////////////////////////////////////////////////////
     tsComment:   case c of
                    '-':  inc(CommentState);
                    '>':  if CommentState=2 then State:=tsContent;
                    else  CommentState:=0;  
                  end;
    ////////////////////////////////////////////////////////////////////////////
    end;
  end;

  Attrs:=Trim(attrs);
  if (length(Attrs)>0) and (attrs[length(Attrs)]='/') then begin
    TagType:=tt_Single;
    setlength(Attrs, length(Attrs)-1);
  end;
end;

procedure TXmlTree.LoadFromStream(const Stream: TStream);
var
  PrevContent, TagName: string;
  TagType: TTagType;
  Node:   TXmlNode;
  Attrs:  string;
begin
  if FRoot<>nil then FRoot.Free;
  FRoot:=nil;
  Node:=nil;
  try
    while GetNextTag(Stream, PrevContent, TagName, Attrs, TagType) do begin
      case TagType of
        //////////////////////////////////////////////////////////////////////////
        tt_Single,
        tt_Open:  begin
                    Node:=TXmlNode.Create(Node);
                    if FRoot=nil then FRoot:=Node;

                    Node.FName:=TagName;
                    ParseAttributes(attrs, Node.Attributes);

                    if Node.Parent<>nil then Node.Parent.Content:=ClearContent(Node.Parent.Content+PrevContent);
                    if TagType=tt_Single then Node:=Node.Parent;
                  end;
        //////////////////////////////////////////////////////////////////////////
        tt_Close: begin
                    if Node=nil then raise Exception.Create(format(XML_UNEXPECTED_CLOSE_TAG,[TagName]));
                    if Node.Name<>TagName then raise Exception.Create(format(XML_BAD_CLOSE_TAG, [Node.Name, TagName]));

                    Node.Content:=ClearContent(Node.Content+PrevContent);
                    Node:=Node.Parent;
                  end;
        //////////////////////////////////////////////////////////////////////////
      end;
    end;
  finally
    if FRoot=nil then FRoot:=TXmlNode.Create(nil);
  end;
end;

procedure TXmlTree.ParseAttributes(S: string; Attributes: TStringList);
type
  TState=(s_name, s_value);
var
  i: integer;
  State: TState;
  name, value: string;
begin
  name:=''; value:='';
  State:=s_name;

  for i:=1 to length(S) do begin
    case State of
      //////////////////////////////////////////////////////////////////////////
      s_name:
              case S[i] of
                #0..' ':  ;
                '=':    State:=s_Value;
                else    name:=name+S[i];
              end;
      //////////////////////////////////////////////////////////////////////////
      s_value:
              case S[i] of
                #0..' ': ;
                '''',
                '"':  if value<>'' then begin
                        state:=s_name;
                        Attributes.Add(name+'='+value);
                        name:='';
                        value:='';
                      end;
                else  value:=value+S[i];
              end;
      //////////////////////////////////////////////////////////////////////////
    end;
  end;
end;

function TXmlTree.ClearContent(S: string):string;
var
  b, e: integer;
begin
  if S='' then begin
    result:='';
    exit;
  end;

  b:=1;
  while (ord(S[b])>0) and (ord(S[b])<=32) do inc(b);

  e:=length(S);
  while (e>0) and (ord(S[e])<=32) do dec(e);

  result:=copy(S, b, e-b+1);
end;

procedure TXmlTree.SaveToStream(const Stream: TStream);
  procedure WriteString(Value: string);
  begin
    Stream.WriteBuffer(Value[1], length(Value));
  end;

  procedure WriteNode(Node: TXmlNode; Indent: string; IsFirst: boolean=false);
  var
    i: integer;
  begin
    if not IsFirst then WriteString(CRLF);
    if pos(' ', Node.Name)>0 then WriteString(Indent+'<"'+Node.Name+'"')
    else WriteString(Indent+'<'+Node.Name);
    
    for i:=0 to Node.Attributes.Count-1 do
      WriteString(' '+Node.Attributes.Names[i]+'='''+Node.Attributes.Values[Node.Attributes.Names[i]]+'''');

    if (Node.Count=0) and (Node.Content='') then begin
      WriteString('/>');
      exit;
    end;

    WriteString('>');
    WriteString(Node.Content);
    for i:=0 to Node.Count-1 do WriteNode(Node[i], Indent+TAB);

    if Node.Count>0 then WriteString(CRLF+Indent);
    if pos(' ', Node.Name)>0 then WriteString('</"'+Node.Name+'">')
    else WriteString('</'+Node.Name+'>');
  end;

begin
  WriteNode(Root, '', True);
end;


procedure TXmlTree.LoadFromFile(const FileName: string);
var
  FStream: TFileStream;
begin
  FStream:=TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(FStream);
  finally
    FStream.Free;
  end;
end;

procedure TXmlTree.SaveToFile(const FileName: string);
var
  FStream: TFileStream;
begin
  FStream:=TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(FStream);
  finally
    FStream.Free;
  end;
end;

procedure TXmlTree.ParsePath(const Path: string; Result: TStrings);
var
  i, n: integer;
begin
  Result.Clear;
  if Path='' then exit;
  n:=0;
  for i:=1 to length(Path)+1 do begin
    case Path[i] of
      #0,
      '/',
      '\':  begin
              if n>0 then Result.Add(copy(Path, i-n, n));
              n:=0;
            end;
      else  inc(n);
    end;
  end;
end;

function TXmlTree.DoGetNode(var Path: TStrings): TXmlNode;
var
  Node: TXmlNode;
begin
  Node:=Root;
  result:=Root;
  while Path.Count>0 do begin
    Node:=Node.NodeByName(Path[0], CaseSensitive);
    if Node=nil then exit;
    result:=Node;
    Path.Delete(0);
  end;
end;

function TXmlTree.Exist(const Path: string): boolean;
var
  strs: TStrings;
begin
  strs:=TStringList.Create;
  ParsePath(Path, strs);
  DoGetNode(strs);
  result:=strs.Count=0;
  strs.Free;
end;

function TXmlTree.ByPath(const APath: string): TXmlNode;
var
  strs: TStrings;
  i: integer;
begin
  strs:=TStringList.Create;
  ParsePath(APath, strs);
  result:=DoGetNode(strs);
  for i:=0 to strs.Count-1 do result:=result.Add(strs[i]);
  strs.Free;
end;

procedure TXmlTree.SetCaseSens(const Value: boolean);
begin
  FRoot.FCaseSens := Value;
end;

function TXmlTree.GetCaseSens: boolean;
begin
  result:=Froot.CaseSensitive;
end;



end.
